"""The demography module contains functions and classes for modelling the demography of
tree communities, including the definition of plant functional types, size structured
plant cohorts, community canopy models and the allocation of gross primary productivity
into respiration, turnover and growth.
"""  # noqa: D205
